#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/io.h>
#include <linux/jiffies.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/pagemap.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/vmalloc.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
	#include <linux/sched/signal.h>
#else
	#include <linux/sched.h>
#endif

#include "ximea_cam_pcie.h"

// compatibility with newer kernels
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0)
	#define __devinit
	#define __devexit
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)
	#define page_cache_release(page) put_page(page)
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,4,168) && LINUX_VERSION_CODE < KERNEL_VERSION(4,5,0)
	#define get_user_pages(a, b, c, d, e, f, g, h) get_user_pages(a, b, c, d, e ? FOLL_WRITE : 0, g, h)
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,6,0)
	#if LINUX_VERSION_CODE < KERNEL_VERSION(4,9,0)
		#define get_user_pages(a, b, c, d, e, f, g, h) get_user_pages(c, d, e, f, g, h)
	#else
		#define get_user_pages(a, b, c, d, e, f, g, h) get_user_pages(c, d, e ? FOLL_WRITE : 0, g, h)
	#endif
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,8,0)
	#undef DEFINE_DMA_ATTRS
	#define DEFINE_DMA_ATTRS(x) unsigned long x = 0
	#undef dma_set_attr
	#define dma_set_attr(a, b) (*(b) |= (a))
	#define dma_map_single_attrs(a, b, c, d, e) dma_map_single_attrs(a, b, c, d, *(e))
	#define dma_map_sg_attrs(a, b, c, d, e) dma_map_sg_attrs(a, b, c, d, *(e))
	#define dma_unmap_single_attrs(a, b, c, d, e) dma_unmap_single_attrs(a, b, c, d, *(e))
	#define dma_unmap_sg_attrs(a, b, c, d, e) dma_unmap_sg_attrs(a, b, c, d, *(e))
#endif

#define DRV_NAME "ximea_cam_pcie"

static unsigned int MAX_NUM_BUF = 300000;
module_param(MAX_NUM_BUF, uint, S_IRUGO);
MODULE_PARM_DESC(MAX_NUM_BUF, "Maximum number of buffers per device");

#define MAX_DEVICES 100
static atomic_t minors[MAX_DEVICES];
static dev_t chrdev;
static struct class *devclass;

struct sh_dev {
	struct pci_dev *pci_dev;
	struct cdev cdev;
	struct page** *mappedpages;
	struct scatterlist **sgs;
	enum dma_data_direction *sg_dir;
	u32 *lockedpages;
	struct semaphore memsem;
};

void unlock_mem(struct sh_dev *sh, unsigned long index) {
	int i;
	DEFINE_DMA_ATTRS(attrs);
	dma_set_attr(DMA_ATTR_SKIP_CPU_SYNC, &attrs);

	dma_unmap_sg_attrs(&sh->pci_dev->dev, sh->sgs[index], sh->lockedpages[index], sh->sg_dir[index], &attrs);
	for(i=0;i<sh->lockedpages[index];i++) {
		if(!PageReserved(sh->mappedpages[index][i]))
			SetPageDirty(sh->mappedpages[index][i]);
		page_cache_release(sh->mappedpages[index][i]);
	}
	vfree(sh->sgs[index]);
	sh->sgs[index] = 0;
	vfree(sh->mappedpages[index]);
	sh->lockedpages[index] = 0;
}

static long ximea_cam_pcie_ioctl(struct file *file, unsigned int cmd, unsigned long arg) {
	struct sh_dev *sh = file->private_data;
	ximea_cam_pcie_lockmem_args lockmem;
	ximea_cam_pcie_unlockmem_args unlockmem;
	int c;
	unsigned int i, j = 0;
	int ret = -ENOTTY;
	struct scatterlist *sg;
	struct vm_area_struct *vma;
	DEFINE_DMA_ATTRS(attrs);
	dma_set_attr(DMA_ATTR_SKIP_CPU_SYNC, &attrs);

	switch(cmd) {
	case XIMEA_CAM_PCIE_IOCLOCKMEM:
		if(down_interruptible(&sh->memsem))
			return -ERESTARTSYS;
		c = copy_from_user(&lockmem, (void*)arg, sizeof(ximea_cam_pcie_lockmem_args));
		for(i=0;i<MAX_NUM_BUF;i++)
			if(!sh->lockedpages[i])
				break;
		if(c != 0 || lockmem.pages <= 0 || i == MAX_NUM_BUF) {
			up(&sh->memsem);
			break;
		}
		unlockmem.index = i;
		unlockmem.pages = (unsigned long)(lockmem.buffer+lockmem.bytes-1)/PAGE_SIZE-(unsigned long)lockmem.buffer/PAGE_SIZE+1;
		if(unlockmem.pages > lockmem.pages) {
			up(&sh->memsem);
			break;
		}
		down_read(&current->mm->mmap_sem);
		vma = find_vma(current->mm, (unsigned long)lockmem.buffer);
		if(!vma || vma->vm_start > (unsigned long)lockmem.buffer) { //invalid address
			up_read(&current->mm->mmap_sem);
			up(&sh->memsem);
			break;
		}
		if(vma->vm_flags & (VM_WRITE | VM_READ)) { //buffer points to a valid CPU memory location
			up_read(&current->mm->mmap_sem);
			unlockmem.mappedpages = vmalloc(sizeof(struct page)*unlockmem.pages);
			if(unlockmem.mappedpages == 0) {
				up(&sh->memsem);
				break;
			}
			sh->sgs[unlockmem.index] = vmalloc(sizeof(*sh->sgs[unlockmem.index]) * unlockmem.pages);
			if(!sh->sgs[unlockmem.index]) {
				vfree(unlockmem.mappedpages);
				up(&sh->memsem);
				break;
			}
			sg_init_table(sh->sgs[unlockmem.index], unlockmem.pages);
			sh->sg_dir[unlockmem.index] = lockmem.write ? DMA_TO_DEVICE : DMA_FROM_DEVICE;
			down_read(&current->mm->mmap_sem);
			c = get_user_pages(current, current->mm, (unsigned long)lockmem.buffer/PAGE_SIZE*PAGE_SIZE, unlockmem.pages, 1, 0, (struct page**)unlockmem.mappedpages, 0);
			up_read(&current->mm->mmap_sem);
			if(c != unlockmem.pages) goto LOCK_CLEANUP;
			j = (unsigned long)lockmem.buffer % PAGE_SIZE; //offset
			for_each_sg(sh->sgs[unlockmem.index], sg, unlockmem.pages, i) {
				sg_set_page(sg, ((struct page**)unlockmem.mappedpages)[i], min(PAGE_SIZE - j, (unsigned long)lockmem.bytes), j);
				lockmem.bytes -= PAGE_SIZE - j;
				j = 0;
			}
			j = dma_map_sg_attrs(&sh->pci_dev->dev, sh->sgs[unlockmem.index], unlockmem.pages, sh->sg_dir[unlockmem.index], &attrs);
			if(!j) {
				printk(KERN_DEBUG DRV_NAME ": dma_map_sg_attrs returned %u\n", j);
				goto LOCK_CLEANUP;
			}
			for_each_sg(sh->sgs[unlockmem.index], sg, j, i) {
				for(unlockmem.addr = sg_dma_address(sg);
						unlockmem.addr < sg_dma_address(sg) + sg_dma_len(sg);
						unlockmem.addr += unlockmem.bytes) {
					unlockmem.bytes = min(((unlockmem.addr >> PAGE_SHIFT) + 1) << PAGE_SHIFT, (u64)sg_dma_address(sg) + sg_dma_len(sg)) - unlockmem.addr;
					if(copy_to_user(lockmem.out++, &unlockmem, sizeof(ximea_cam_pcie_unlockmem_args)) != 0) goto LOCK_CLEANUP;
				}
			}
		} else { //assume GPU memory
			up_read(&current->mm->mmap_sem);
			up(&sh->memsem);
			ret = -EOPNOTSUPP;
			break;
		}
		sh->lockedpages[unlockmem.index] = unlockmem.pages;
		sh->mappedpages[unlockmem.index] = (struct page**)unlockmem.mappedpages;
		up(&sh->memsem);
		ret = 0;
		break;
LOCK_CLEANUP:
		if(j)
			dma_unmap_sg_attrs(&sh->pci_dev->dev, sh->sgs[unlockmem.index], unlockmem.pages, sh->sg_dir[unlockmem.index], &attrs);
		for(;c>0;c--)
			page_cache_release(((struct page**)unlockmem.mappedpages)[c-1]);
		vfree(sh->sgs[unlockmem.index]);
		sh->sgs[unlockmem.index] = 0;
		vfree(unlockmem.mappedpages);
		up(&sh->memsem);
		break;
	case XIMEA_CAM_PCIE_IOCUNLOCKMEM:
		if(down_interruptible(&sh->memsem))
			return -ERESTARTSYS;
		c = copy_from_user(&unlockmem, (void*)arg, sizeof(ximea_cam_pcie_unlockmem_args));
		if(c != 0 || unlockmem.index >= MAX_NUM_BUF || !sh->lockedpages[unlockmem.index] || unlockmem.pages != sh->lockedpages[unlockmem.index] || unlockmem.mappedpages != sh->mappedpages[unlockmem.index]) {
			up(&sh->memsem);
			break;
		}
		unlock_mem(sh, unlockmem.index);
		up(&sh->memsem);
		ret = 0;
		break;
	case XIMEA_CAM_PCIE_IOCSYNCMEM4CPU:
		if(down_interruptible(&sh->memsem))
			return -ERESTARTSYS;
		c = copy_from_user(&unlockmem, (void*)arg, sizeof(ximea_cam_pcie_unlockmem_args));
		if(c != 0 || unlockmem.index >= MAX_NUM_BUF || !sh->lockedpages[unlockmem.index] || unlockmem.pages != sh->lockedpages[unlockmem.index] || unlockmem.mappedpages != sh->mappedpages[unlockmem.index]) {
			up(&sh->memsem);
			break;
		}
		dma_sync_sg_for_cpu(&sh->pci_dev->dev, sh->sgs[unlockmem.index], unlockmem.pages, sh->sg_dir[unlockmem.index]);
		up(&sh->memsem);
		ret = 0;
		break;
	case XIMEA_CAM_PCIE_IOCSYNCMEM4DEV:
		if(down_interruptible(&sh->memsem))
			return -ERESTARTSYS;
		c = copy_from_user(&unlockmem, (void*)arg, sizeof(ximea_cam_pcie_unlockmem_args));
		if(c != 0 || unlockmem.index >= MAX_NUM_BUF || !sh->lockedpages[unlockmem.index] || unlockmem.pages != sh->lockedpages[unlockmem.index] || unlockmem.mappedpages != sh->mappedpages[unlockmem.index]) {
			up(&sh->memsem);
			break;
		}
		dma_sync_sg_for_device(&sh->pci_dev->dev, sh->sgs[unlockmem.index], unlockmem.pages, sh->sg_dir[unlockmem.index]);
		up(&sh->memsem);
		ret = 0;
		break;
	}
	return ret;
}

static int ximea_cam_pcie_open(struct inode *inode, struct file *filep)
{
	struct sh_dev *sh = container_of(inode->i_cdev, struct sh_dev, cdev);
	filep->private_data = sh;
	return 0;
}

static int ximea_cam_pcie_release(struct inode *inode, struct file *filep)
{
	struct sh_dev *sh = filep->private_data;
	int i;
	for(i=0;i<MAX_NUM_BUF;i++) {
		if(sh->lockedpages[i]) {
			//printk(KERN_DEBUG DRV_NAME ": automatically unlocking memory on device release\n");
			unlock_mem(sh, i);
		}
	}
	return 0;
}

static struct file_operations ximea_cam_pcie_fops = {
	.owner			= THIS_MODULE,
	.unlocked_ioctl	= ximea_cam_pcie_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl	= ximea_cam_pcie_ioctl,
#endif
	.open			= ximea_cam_pcie_open,
	.release		= ximea_cam_pcie_release,
};

static int __devinit probe(struct pci_dev *dev, const struct pci_device_id *id)
{
	int minor, rc;
	struct sh_dev *sh = NULL;
	struct device *sysdev;

	printk(KERN_DEBUG "probe(dev = 0x%p, pciid = 0x%p)\n", dev, id);
	for(minor = 0; minor < MAX_DEVICES; minor++) {
		if(atomic_dec_and_test(&minors[minor]))
			break;
		atomic_inc(&minors[minor]);
	}
	if(minor == MAX_DEVICES) {
		printk(KERN_DEBUG "Maximum number of " DRV_NAME " devices (%d) exceeded, ignoring this one.\n", MAX_DEVICES);
		rc = -ENOMEM;
		goto err_minor;
	}
	/* allocate memory for per-board book keeping */
	sh = kzalloc(sizeof(struct sh_dev), GFP_KERNEL);
	if (!sh) {
		printk(KERN_DEBUG "Could not kzalloc()ate memory.\n");
		rc = -ENOMEM;
		goto err_sh;
	}
	sh->mappedpages = vzalloc(MAX_NUM_BUF * sizeof(*sh->mappedpages));
	sh->lockedpages = vzalloc(MAX_NUM_BUF * sizeof(*sh->lockedpages));
	sh->sgs			= vzalloc(MAX_NUM_BUF * sizeof(*sh->sgs        ));
	sh->sg_dir		= vzalloc(MAX_NUM_BUF * sizeof(*sh->sg_dir     ));
	if (!sh->mappedpages || !sh->lockedpages || !sh->sgs || !sh->sg_dir) {
		printk(KERN_DEBUG "Could not vzalloc()ate memory.\n");
		rc = -ENOMEM;
		goto err_enable;
	}
	sh->pci_dev = dev;
	sema_init(&sh->memsem, 1);
	dev_set_drvdata(&dev->dev, (void *)sh);
	printk(KERN_DEBUG "probe() sh = 0x%p\n", sh);
	/* enable device */
	rc = pci_enable_device(dev);
	if (rc) {
		printk(KERN_DEBUG "pci_enable_device() failed\n");
		goto err_enable;
	}
	/* enable bus master capability on device */
	pci_set_master(dev);
	rc = pci_request_regions(dev, DRV_NAME);
	/* could not request all regions? */
	if (rc) {
		/* assume device is in use (and do not disable it later!) */
		goto err_regions;
	}
	/* @see Documentation/DMA-mapping.txt */
	if (!pci_set_dma_mask(dev, DMA_BIT_MASK(64))) {
		pci_set_consistent_dma_mask(dev, DMA_BIT_MASK(64));
		/* use 64-bit DMA */
		printk(KERN_DEBUG "Using a 64-bit DMA mask.\n");
	} else
		if (!pci_set_dma_mask(dev, DMA_BIT_MASK(32))) {
			printk(KERN_DEBUG "Could not set 64-bit DMA mask.\n");
			pci_set_consistent_dma_mask(dev, DMA_BIT_MASK(32));
			/* use 32-bit DMA */
			printk(KERN_DEBUG "Using a 32-bit DMA mask.\n");
		} else {
			printk(KERN_DEBUG "No suitable DMA possible.\n");
			/** @todo Choose proper error return code */
			rc = -1;
			goto err_mask;
		}
	cdev_init(&sh->cdev, &ximea_cam_pcie_fops);
	sh->cdev.owner = THIS_MODULE;
	rc = cdev_add(&sh->cdev, MKDEV(MAJOR(chrdev), minor), 1);
	if(rc < 0) {
		printk(KERN_DEBUG "Char dev creation failed\n");
		goto err_cdev;
	}
	sysdev = device_create(devclass, &dev->dev, sh->cdev.dev, NULL, "ximea%02d", minor);
	if(IS_ERR(sysdev)) {
		rc = PTR_ERR(sysdev);
		printk(KERN_DEBUG "Sys device creation failed\n");
		goto err_dev;
	}
	/* succesfully took the device */
	rc = 0;
	printk(KERN_DEBUG "probe() successful.\n");
	goto end;
err_dev:
	cdev_del(&sh->cdev);
err_cdev:
err_mask:
	/* disable the device if it is not in use */
	pci_disable_device(dev);
	pci_release_regions(dev);
err_regions:
err_enable:
	dev_set_drvdata(&dev->dev, NULL);
	vfree(sh->sg_dir);
	vfree(sh->sgs);
	vfree(sh->lockedpages);
	vfree(sh->mappedpages);
	kfree(sh);
err_sh:
	atomic_inc(&minors[minor]);
err_minor:
end:
	return rc;
}

static void __devexit remove(struct pci_dev *dev)
{
	struct sh_dev *sh;
	int minor;

	printk(KERN_DEBUG "remove(0x%p)\n", dev);
	if ((dev == 0) || (dev_get_drvdata(&dev->dev) == 0)) {
		printk(KERN_DEBUG "remove(dev = 0x%p) dev->dev.driver_data = 0x%p\n", dev, dev_get_drvdata(&dev->dev));
		return;
	}
	sh = (struct sh_dev *)dev_get_drvdata(&dev->dev);
	printk(KERN_DEBUG "remove(dev = 0x%p) where dev->dev.driver_data = 0x%p\n", dev, sh);
	if (sh->pci_dev != dev) {
		printk(KERN_DEBUG "dev->dev.driver_data->pci_dev (0x%08lx) != dev (0x%08lx)\n",
				(unsigned long)sh->pci_dev, (unsigned long)dev);
	}
	/* remove character device */
	minor = MINOR(sh->cdev.dev);
	device_destroy(devclass, sh->cdev.dev);
	cdev_del(&sh->cdev);
	pci_disable_device(dev);
	/* to be called after device disable */
	pci_release_regions(dev);
	dev_set_drvdata(&dev->dev, NULL);
	vfree(sh->sg_dir);
	vfree(sh->sgs);
	vfree(sh->lockedpages);
	vfree(sh->mappedpages);
	kfree(sh);
	atomic_inc(&minors[minor]);
}

static const struct pci_device_id ids[] = {
	{ PCI_DEVICE(0x10de, 0x10e5), },
	{ 0, }
};
MODULE_DEVICE_TABLE(pci, ids);

static struct pci_driver pci_driver = {
	.name = DRV_NAME,
	.id_table = ids,
	.probe = probe,
	.remove = remove,
};

static int __init ximea_cam_pcie_init(void)
{
	int i, rc;

	printk(KERN_DEBUG DRV_NAME " init()\n");
	rc = alloc_chrdev_region(&chrdev, 0, MAX_DEVICES, DRV_NAME);
	if(rc < 0) {
		printk(KERN_DEBUG "Char dev allocation failed\n");
		goto exit_chrdev;
	}
	devclass = class_create(THIS_MODULE, DRV_NAME);
	if(IS_ERR(devclass)) {
		printk(KERN_DEBUG "Class creation failed\n");
		rc = PTR_ERR(devclass);
		goto exit_class;
	}
	for(i = 0; i < MAX_DEVICES; i++)
		atomic_set(&minors[i], 1);
	rc = pci_register_driver(&pci_driver);
	if (rc < 0)
		goto exit_pci;
	rc = 0;
	goto exit_norm;
exit_pci:
	class_destroy(devclass);
exit_class:
	unregister_chrdev_region(chrdev, MAX_DEVICES);
exit_chrdev:
exit_norm:
	return rc;
}

static void __exit ximea_cam_pcie_exit(void)
{
	printk(KERN_DEBUG DRV_NAME " exit()\n");
	/* unregister this driver from the PCI bus driver */
	pci_unregister_driver(&pci_driver);
	class_destroy(devclass);
	unregister_chrdev_region(chrdev, MAX_DEVICES);
}

MODULE_LICENSE("GPL");

module_init(ximea_cam_pcie_init);
module_exit(ximea_cam_pcie_exit);
